/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: TaskProcessor.java 154 6 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management;

import java.util.ArrayList;
import java.util.List;

import org.ow2.petals.util.LoggingUtil;

/**
 * The TaskProcessor executes an ordered List of Tasks : it calls the
 * <i>execute</i> method on each Task. It ensures the system state recovery in
 * case of failure during the Tasks processing : it will call the <i>undo</i>
 * method on each allready executed Task before the whole process crash.
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class TaskProcessor {

    protected List<Task> taskList;

    protected Context context;

    protected LoggingUtil log;

    public TaskProcessor(Context context, LoggingUtil log) {
        super();
        this.taskList = new ArrayList<Task>();
        this.context = context;
        this.log = log;
    }

    /**
     * Add a new Task in the List of Tasks to execute. Task will be processed in
     * the same order as they have been added to the List.
     * 
     * @param task
     *            a new Task to process
     */
    public void addTask(Task task) {
        this.taskList.add(task);
    }

    /**
     * Process each Task previously added to the Processor, in the same order as
     * they have been added to it. If an exception occurs during the processing
     * of a Task, the processor will undo this Task and all previously executed
     * Tasks, in the reverse order as they have been executed.
     * 
     * @return The cause of the error if the process has failed, {@code null}
     *         otherwise
     * 
     */
    public Exception process() {
        Exception result = null;
        for (int i = 0; (i < this.taskList.size()) && (result == null); i++) {
            try {
                this.taskList.get(i).execute(this.context);
            } catch (Exception e) {
                result = e;
                this.unprocess(i - 1);
            }
        }
        return result;
    }

    /**
     * Unprocess each Task previously added to the Processor from the step
     * provided, in the reversed order they have been added to it. If an error
     * occurs during a Task unprocessing, the processor skips it and continues
     * to unprocess other Tasks.
     * 
     * @param step
     *            . The task number reference into the list to start the reverse
     *            processing
     * @return {@code true} if the process is successful, {@code false}
     *         otherwise
     */
    public boolean unprocess(int step) {
        boolean result = true;
        for (int i = step; i >= 0; i--) {
            try {
                this.taskList.get(i).undo(this.context);
            } catch (Exception e) {
                result = false;
                this.log.error("Error during task unprocessing", e);
            }
        }
        return result;
    }

    /**
     * Unprocess each Task previously added to the Processor, in the reversed
     * order they have been added to it. If an error occurs during a Task
     * unprocessing, the processor skips it and continues to unprocess other
     * Tasks.
     * 
     * @return {@code true} if the process is successful, {@code false}
     *         otherwise
     */
    public boolean unprocess() {
        return this.unprocess(this.taskList.size() - 1);
    }
}
