/**
 * PETALS: PETALS Services Platform Copyright (C) 2005 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: MessageExchangeFactoryImpl.java,v 1.3 2005/07/25 07:44:35 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.exchange;

import java.net.URI;

import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.RobustInOnly;
import javax.jbi.messaging.MessageExchange.Role;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.util.monolog.api.Logger;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation;
import org.ow2.petals.commons.util.IDGenerator;
import org.ow2.petals.jbi.messaging.exchange.types.InOnlyImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOptionalOutImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOutImpl;
import org.ow2.petals.jbi.messaging.exchange.types.RobustInOnlyImpl;
import org.ow2.petals.util.LoggingUtil;

/**
 * @author Adrien LOUIS - EBM WebSourcing
 * 
 */
public class MessageExchangeFactoryImpl implements MessageExchangeFactory {

    /**
     * The bound Delivery Channel
     */
    protected DeliveryChannel deliveryChannel;

    /**
     * The consumer endpoint, the return address for any service request
     */
    protected org.ow2.petals.jbi.messaging.endpoint.ServiceEndpoint consumerEndpoint;

    /**
     * The default target Interface if set
     */
    protected QName defaultInterfaceName;

    /**
     * The default target Service if set
     */
    protected QName defaultServiceName;

    /**
     * The default target Endpoint if set
     */
    protected ServiceEndpoint defaultEndpoint;

    /**
     * Logger wrapper.
     */
    protected LoggingUtil log;

    public MessageExchangeFactoryImpl(final DeliveryChannel deliveryChannel,
            final org.ow2.petals.jbi.messaging.endpoint.ServiceEndpoint consumerEndpoint,
            final Logger logger) {
        this.log = new LoggingUtil(logger);
        this.deliveryChannel = deliveryChannel;
        this.consumerEndpoint = consumerEndpoint;
    }

    public MessageExchangeFactoryImpl(final DeliveryChannel deliveryChannel,
            final org.ow2.petals.jbi.messaging.endpoint.ServiceEndpoint consumerEndpoint,
            final Logger logger, QName interfaceName, QName serviceName, ServiceEndpoint endpoint) {

        this(deliveryChannel, consumerEndpoint, logger);
        this.defaultInterfaceName = interfaceName;
        this.defaultServiceName = serviceName;
        this.defaultEndpoint = endpoint;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.jbi.messaging.MessageExchangeFactory#createExchange(javax.xml.namespace
     * .QName, javax.xml.namespace.QName)
     */
    public MessageExchange createExchange(QName serviceName, QName operationName)
            throws MessagingException {
        this.log.call("Service Name : " + serviceName + " - operationName : " + operationName);

        MessageExchangeImpl messageExchange = this.createExchange();

        this.setDefaultMessageExchangeProperties(messageExchange);
        messageExchange.setService(serviceName);
        messageExchange.setOperation(operationName);

        return new MessageExchangeDecorator(messageExchange, Role.CONSUMER);
    }

    /**
     * Return a <code>MessageExchangeDecorator</code> for Consumer, as
     * <code>InOnlyExchange, InOutExchange,...</code>
     * 
     * @see javax.jbi.messaging.MessageExchangeFactory#createExchange(java.net.URI)
     */
    public MessageExchange createExchange(URI pattern) throws MessagingException {
        this.log.call(pattern);

        if (!MessageExchangeImpl.IN_ONLY_PATTERN.equals(pattern)
                && !MessageExchangeImpl.ROBUST_IN_ONLY_PATTERN.equals(pattern)
                && !MessageExchangeImpl.IN_OUT_PATTERN.equals(pattern)
                && !MessageExchangeImpl.IN_OPTIONAL_OUT_PATTERN.equals(pattern)
                && !AbsItfOperation.MEPPatternConstants.IN_ONLY.equals(pattern)
                && !AbsItfOperation.MEPPatternConstants.ROBUST_IN_ONLY.equals(pattern)
                && !AbsItfOperation.MEPPatternConstants.IN_OUT.equals(pattern)
                && !AbsItfOperation.MEPPatternConstants.IN_OPTIONAL_OUT.equals(pattern)) {
            throw new MessagingException("This Message Exchange Pattern is not recognized by JBI.");
        }

        return this.createExchangeDecorator(this.createExchange(), pattern);
    }

    /**
     * Return a <code>MessageExchangeDecorator</code>, as
     * <code>InOnlyExchange, InOutExchange,...</code>, or
     * <code>MessageExchangeDecorator</code> if pattern is unknown. The
     * 'observer' role is set with the current MessageExchangeImpl.getRole()
     * property.
     * 
     * @see javax.jbi.messaging.MessageExchangeFactory#createExchange(java.net.URI)
     */
    public MessageExchange createExchangeDecorator(MessageExchangeImpl exchange, URI pattern)
            throws MessagingException {
        this.log.call(pattern);

        if (MessageExchangeImpl.IN_ONLY_PATTERN.equals(pattern)
                || AbsItfOperation.MEPPatternConstants.IN_ONLY.equals(pattern)) {
            return new InOnlyImpl(exchange, exchange.getRole());
        } else if (MessageExchangeImpl.ROBUST_IN_ONLY_PATTERN.equals(pattern)
                || AbsItfOperation.MEPPatternConstants.ROBUST_IN_ONLY.equals(pattern)) {
            return new RobustInOnlyImpl(exchange, exchange.getRole());
        } else if (MessageExchangeImpl.IN_OUT_PATTERN.equals(pattern)
                || AbsItfOperation.MEPPatternConstants.IN_OUT.equals(pattern)) {
            return new InOutImpl(exchange, exchange.getRole());
        } else if (MessageExchangeImpl.IN_OPTIONAL_OUT_PATTERN.equals(pattern)
                || AbsItfOperation.MEPPatternConstants.IN_OPTIONAL_OUT.equals(pattern)) {
            return new InOptionalOutImpl(exchange, exchange.getRole());
        } else {
            return new MessageExchangeDecorator(exchange, exchange.getRole());
        }
    }

    /**
     * Return a <code>MessageExchangeDecorator</code> for Consumer, as
     * <code>InOnly</code>.
     * 
     * @see javax.jbi.messaging.MessageExchangeFactory#createInOnlyExchange()
     */
    public InOnly createInOnlyExchange() throws MessagingException {
        this.log.call();

        MessageExchangeImpl msg = this.createExchange();

        return new InOnlyImpl(msg, Role.CONSUMER);
    }

    /**
     * Return a <code>MessageExchangeDecorator</code> for Consumer, as
     * <code>InOptionalOut</code>.
     * 
     * @see javax.jbi.messaging.MessageExchangeFactory#createInOptionalOutExchange()
     */
    public InOptionalOut createInOptionalOutExchange() throws MessagingException {
        this.log.call();

        MessageExchangeImpl msg = this.createExchange();

        return new InOptionalOutImpl(msg, Role.CONSUMER);
    }

    /**
     * Return a <code>MessageExchangeDecorator</code> for Consumer, as
     * <code>InOut</code>.
     * 
     * @see javax.jbi.messaging.MessageExchangeFactory#createInOutExchange()
     */
    public InOut createInOutExchange() throws MessagingException {
        this.log.call();

        MessageExchangeImpl msg = this.createExchange();

        return new InOutImpl(msg, Role.CONSUMER);
    }

    /**
     * Return a <code>MessageExchangeDecorator</code> for Consumer, as
     * <code>RobustInOnly</code>.
     * 
     * @see javax.jbi.messaging.MessageExchangeFactory#createRobustInOnlyExchange()
     */
    public RobustInOnly createRobustInOnlyExchange() throws MessagingException {
        this.log.call();

        MessageExchangeImpl msg = this.createExchange();

        return new RobustInOnlyImpl(msg, Role.CONSUMER);
    }

    /**
     * Create a MessageExchangeImpl, which is the real object that is exchange
     * between the components
     * 
     * @return
     * @throws MessagingException
     */
    private final MessageExchangeImpl createExchange() throws MessagingException {
        this.log.call();

        this.deliveryChannel.checkDeliveryChannelIsOpened();

        MessageExchangeImpl msg = new MessageExchangeImpl(this.consumerEndpoint);
        this.setDefaultMessageExchangeProperties(msg);
        msg.setExchangeId(IDGenerator.getInstance().getNewID());

        return msg;
    }

    // //////////////////////////////////////////////////////
    // specific methods
    // //////////////////////////////////////////////////////

    /**
     * set to the MessageExchangeImpl the defaultInterfaceName,
     * defaultServiceName and defaultServiceEndpoint, if they were specified
     * during the creation of this factory.
     * 
     * @param msg
     */
    private final void setDefaultMessageExchangeProperties(MessageExchangeImpl msg) {
        this.log.call();

        if (this.defaultInterfaceName != null) {
            msg.setInterfaceName(this.defaultInterfaceName);
        }
        if (this.defaultServiceName != null) {
            msg.setService(this.defaultServiceName);
        }
        if (this.defaultEndpoint != null) {
            msg.setEndpoint(this.defaultEndpoint);
        }

    }
}
