/**
 * PETALS - PETALS Services Platform. Copyright (c) 2009 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */

package org.ow2.petals.system.persistence;

import java.io.IOException;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;

/**
 * The Persistent Service.
 * 
 * @author rnaudin
 * @since 3.1
 */
public interface PersistenceService {

    /**
     * Default Persistence duration
     */
    static final long DEFAULT_DURATION = 3600000;

    /**
     * The default Password configuration
     */
    static final String DEFAULT_PASSWORD = "";

    /**
     * The default URL configuration (File database)
     */
    static final String DEFAULT_URL_PREFIX = "jdbc:hsqldb:file:";

    static final String DEFAULT_URL_SUFFIX = "persistenceService;shutdown=true";

    /**
     * The default User configuration
     */
    static final String DEFAULT_USER = "SA";

    /**
     * The HSQLDB driver
     */
    static final String HSQLDB_DRIVER = "org.hsqldb.jdbcDriver";
    
    /**
     * The ORACLE driver
     */
    static final String ORACLE_DRIVER = "oracle.jdbc.driver.OracleDriver";

    /**
     * The MYSQL driver
     */
    static final String MYSQL_DRIVER = "com.mysql.jdbc.Driver";

    /**
     * Add a Message Exchange to the persistence service.
     * 
     * @param exchange
     *            the exchange
     * @param storeExchange
     *            if {@code true}, persist the body of the exchange. If {@code
     *            false}, persist only the exchangeId and the current timestamp
     * @param isSend
     *            if {@code true}, it is an exchange sent<br>
     *            if {@code false}, it is an exchange receipt if {@code null},
     *            it is nor a send or receipt
     * @throws SQLException
     * @throws IOException
     */
    void addMessageExchange(MessageExchangeImpl exchange, boolean storeExchange, Boolean isSent)
            throws SQLException, IOException;

    /**
     * Delete the exchange.
     * 
     * @param exchangeId
     */
    void deleteMessageExchange(String exchangeId) throws SQLException;

    /**
     * Retrieve a Message Exchange from the persistence.
     * 
     * @param exchangeId
     *            the Id of the exchange
     * 
     * @return the recreated {@code MessageExchangeImpl} instance
     * 
     * @throws SQLException
     * @throws IOException
     * @throws ClassNotFoundException
     */
    MessageExchangeImpl getMessageExchange(String exchangeId) throws SQLException, IOException,
            ClassNotFoundException;

    /**
     * Get the Message Exchange sent/received history from the persistence.
     * 
     * @param exchangeId
     *            the Id of the exchange
     * 
     * @return a map of events, the type (sent or received) as key, the
     *         timestamp in millisecond as value.
     * 
     * @throws SQLException
     */
    Map<String, Long> getMessageExchangeHistory(String exchangeId) throws SQLException;

    /**
     * Retrieve a list of Message Exchange Ids from the persistence matching the
     * given criteria.
     * 
     * @param exchangeId
     *            the Id of the exchange
     */
    List<String> getMessageExchangeIds(QName interfaceName, QName serviceName, String endpointName,
            QName operationName, long startDate, long endDate) throws SQLException;

    /**
     * Count the number of processed exchanges matching the given criteria.
     * 
     * @param exchangeId
     *            the Id of the exchange
     */
    int getMessageExchanges(QName interfaceName, QName serviceName, String endpointName,
            QName operationName, long startDate, long endDate) throws SQLException;

    /**
     * Set the exchange storage duration. When the duration is reached, the
     * exchanges are removed from the Persistence Service.
     * 
     * @param duration
     */
    void setExchangeStorageDuration(long duration);

    /**
     * Update a Message Exchange in the persistence service.
     * 
     * @param exchange
     *            the exchange
     * @param storeExchange
     *            if {@code true}, persist the body of the exchange. If {@code
     *            false}, persist only the exchangeId and the current timestamp
     * @param isSend
     *            if {@code true}, it is an exchange sent<br>
     *            if {@code false}, it is an exchange receipt if {@code null},
     *            it is nor a send or receipt
     * @throws SQLException
     * @throws IOException
     */
    void updateMessageExchange(MessageExchangeImpl exchange, boolean storeExchange, Boolean isSent)
            throws SQLException, IOException;

}
